@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 2007-2014 by                                                  */
+/*  Copyright (C) 2007-2019 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -19,8 +19,7 @@
 
 #include FT_OUTLINE_H
 #include FT_LCD_FILTER_H
-#include FT_CFF_DRIVER_H
-#include FT_TRUETYPE_DRIVER_H
+#include FT_DRIVER_H
 
   /* showing driver name -- the two internal header files */
   /* shouldn't be used in normal programs                 */
@@ -53,11 +52,8 @@
       "            `.afm' or `.pfm').\n"
       "\n" );
     fprintf( stderr,
-      "  -w W         Set the window width to W pixels (default: %dpx).\n"
-      "  -h H         Set the window height to H pixels (default: %dpx).\n"
-      "\n",
-             DIM_X, DIM_Y );
-    fprintf( stderr,
+      "  -w W         Set the window width to W pixels (default: 640px).\n"
+      "  -h H         Set the window height to H pixels (default: 480px).\n"
       "  -r R         Use resolution R dpi (default: 72dpi).\n"
       "  -s S         Set character size to S points (default: 16pt).\n"
       "  -f TEXTFILE  Change displayed text, using text in TEXTFILE\n"
@@ -171,8 +167,6 @@
     "native hinter"
   };
 
-#define HINTING_ENGINE_MAX  2
-
   /** RENDER STATE **/
 
   typedef struct  ColumnStateRec_
@@ -190,6 +184,8 @@
     int            fw_index;
 
     unsigned int   cff_hinting_engine;
+    unsigned int   type1_hinting_engine;
+    unsigned int   t1cid_hinting_engine;
     unsigned int   tt_interpreter_versions[3];
     int            num_tt_interpreter_versions;
     int            tt_interpreter_version_idx;
@@ -236,6 +232,8 @@
                      FT_Library   library )
   {
     FT_UInt  cff_hinting_engine;
+    FT_UInt  type1_hinting_engine;
+    FT_UInt  t1cid_hinting_engine;
     FT_Bool  warping;
 
     unsigned int  tt_interpreter_versions[3]  = { 0, 0, 0 };
@@ -265,6 +263,12 @@
     FT_Property_Get( library,
                      "cff",
                      "hinting-engine", &cff_hinting_engine );
+    FT_Property_Get( library,
+                     "type1",
+                     "hinting-engine", &type1_hinting_engine );
+    FT_Property_Get( library,
+                     "t1cid",
+                     "hinting-engine", &t1cid_hinting_engine );
 
     /* collect all available versions, then set again the default */
     FT_Property_Get( library,
@@ -295,6 +299,8 @@
     state->columns[0].lcd_filter             = FT_LCD_FILTER_DEFAULT;
     state->columns[0].hint_mode              = HINT_MODE_BYTECODE;
     state->columns[0].cff_hinting_engine     = cff_hinting_engine;
+    state->columns[0].type1_hinting_engine   = type1_hinting_engine;
+    state->columns[0].t1cid_hinting_engine   = t1cid_hinting_engine;
 
     state->columns[0].tt_interpreter_versions[0] =
       tt_interpreter_versions[0];
@@ -310,7 +316,7 @@
     state->columns[0].warping                = warping;
     state->columns[0].use_custom_lcd_filter  = 0;
     state->columns[0].fw_index               = 2;
-    /* freetype default filter weights */
+    /* FreeType default filter weights */
     memcpy( state->columns[0].filter_weights, "\x08\x4D\x56\x4D\x08", 5 );
 
     state->columns[1]                        = state->columns[0];
@@ -579,6 +585,15 @@
                      "cff",
                      "hinting-engine",
                      &column->cff_hinting_engine );
+    FT_Property_Set( state->library,
+                     "type1",
+                     "hinting-engine",
+                     &column->type1_hinting_engine );
+    FT_Property_Set( state->library,
+                     "t1cid",
+                     "hinting-engine",
+                     &column->t1cid_hinting_engine );
+
     FT_Property_Set( state->library,
                      "truetype",
                      "interpreter-version",
@@ -703,9 +718,9 @@
       if ( rmode != HINT_MODE_AUTOHINT_LIGHT_SUBPIXEL &&
            column->use_deltas                         )
       {
-        if ( prev_rsb_delta - face->glyph->lsb_delta >= 32 )
+        if ( prev_rsb_delta - face->glyph->lsb_delta > 32 )
           x_origin -= 64;
-        else if ( prev_rsb_delta - face->glyph->lsb_delta < -32 )
+        else if ( prev_rsb_delta - face->glyph->lsb_delta < -31 )
           x_origin += 64;
       }
       prev_rsb_delta = face->glyph->rsb_delta;
@@ -806,15 +821,41 @@
         {
           switch ( column->cff_hinting_engine )
           {
-          case FT_CFF_HINTING_FREETYPE:
+          case FT_HINTING_FREETYPE:
             extra = " (CFF FT)";
             break;
-          case FT_CFF_HINTING_ADOBE:
+          case FT_HINTING_ADOBE:
             extra = " (CFF Adobe)";
             break;
           }
         }
 
+        else if ( !strcmp( module->clazz->module_name, "type1" ) )
+        {
+          switch ( column->type1_hinting_engine )
+          {
+          case FT_HINTING_FREETYPE:
+            extra = " (T1 FT)";
+            break;
+          case FT_HINTING_ADOBE:
+            extra = " (T1 Adobe)";
+            break;
+          }
+        }
+
+        else if ( !strcmp( module->clazz->module_name, "t1cid" ) )
+        {
+          switch ( column->t1cid_hinting_engine )
+          {
+          case FT_HINTING_FREETYPE:
+            extra = " (CID FT)";
+            break;
+          case FT_HINTING_ADOBE:
+            extra = " (CID Adobe)";
+            break;
+          }
+        }
+
         else if ( !strcmp( module->clazz->module_name, "truetype" ) )
         {
           switch ( column->tt_interpreter_versions[
@@ -832,7 +873,7 @@
           }
         }
       }
-      else if ( rmode == HINT_MODE_AUTOHINT_LIGHT )
+      else if ( rmode == HINT_MODE_AUTOHINT )
         extra = warping ? " (+warp)" : " (-warp)";
 
       sprintf( temp, "%s%s",
@@ -1116,36 +1157,35 @@
 
     grWriteln( buf );
     grLn();
-    grWriteln( "This program displays text using various hinting algorithms." );
-    grLn();
     grWriteln( "Use the following keys:" );
     grLn();
-    grWriteln( "  F1, ?        display this help screen" );
+    /*          |----------------------------------|    |----------------------------------| */
+    grWriteln( "F1, ?       display this help screen                                        " );
+    grWriteln( "                                                                            " );
+    grWriteln( "global parameters:                                                          " );
+    grWriteln( "  p, n        previous/next font        1, 2, 3      select column          " );
+    grWriteln( "  Up, Down    adjust size by 0.5pt      Left, Right  switch between columns " );
+    grWriteln( "  PgUp, PgDn  adjust size by 5pt        g, v         adjust gamma value     " );
+    grWriteln( "                                                                            " );
+    grWriteln( " per-column parameters:                                                     " );
+    grWriteln( "  d           toggle lsb/rsb deltas     hinting modes:                      " );
+    grWriteln( "  h           cycle hinting mode          A          unhinted               " );
+    grWriteln( "  H           cycle hinting engine        B          auto-hinter            " );
+    grWriteln( "               (if CFF or TTF)            C          light auto-hinter      " );
+    grWriteln( "  w           toggle warping (if          D          light auto-hinter      " );
+    grWriteln( "               normal auto-hinting)                   (subpixel)            " );
+    grWriteln( "  k           toggle kerning (only        E          native hinter          " );
+    grWriteln( "               from `kern' table)                                           " );
+    grWriteln( "  r           toggle rendering mode                                         " );
+    grWriteln( "  x           toggle layout mode                                            " );
+    grWriteln( "                                                                            " );
+    grWriteln( "  l           cycle LCD filtering                                           " );
+    grWriteln( "  [, ]        select custom LCD                                             " );
+    grWriteln( "               filter weight                                                " );
+    grWriteln( "  -, +(=)     adjust selected custom                                        " );
+    grWriteln( "               LCD filter weight                                            " );
+    /*          |----------------------------------|    |----------------------------------| */
     grLn();
-    grWriteln( "  p, n         select previous/next font" );
-    grLn();
-    grWriteln( " global parameters:" );
-    grLn();
-    grWriteln( "  Up, Down     adjust pointsize by 0.5 unit" );
-    grWriteln( "  PgUp, PgDn   adjust pointsize by 5 units" );
-    grWriteln( "  g, v         adjust gamma value" );
-    grLn();
-    grWriteln( "  1, 2, 3      select left, middle, or right column" );
-    grWriteln( "  Left, Right  switch between columns" );
-    grLn();
-    grWriteln( " per-column parameters:" );
-    grLn();
-    grWriteln( "  d            toggle lsb/rsb deltas" );
-    grWriteln( "  h            cycle hinting mode" );
-    grWriteln( "  H            cycle hinting engine (if CFF or TTF)" );
-    grWriteln( "  w            toggle warping (if light auto-hinting" );
-    grWriteln( "  k            toggle kerning (only from `kern' table)" );
-    grWriteln( "  r            toggle rendering mode" );
-    grWriteln( "  x            toggle layout mode" );
-    grLn();
-    grWriteln( "  l            change LCD filter type" );
-    grWriteln( "  [, ]         select custom LCD filter weight" );
-    grWriteln( "  -, +(=)      adjust selected custom LCD filter weight");
     grLn();
     grWriteln( "press any key to exit this help screen" );
 
@@ -1204,6 +1244,13 @@
     ColumnState  column = &state->columns[state->col];
 
 
+    if ( event->key >= 'A'                &&
+         event->key < 'A' + HINT_MODE_MAX )
+    {
+      column->hint_mode = (HintMode)( event->key - 'A' );
+      return ret;
+    }
+
     switch ( event->key )
     {
     case grKeyEsc:
@@ -1276,18 +1323,24 @@
         {
           if ( !strcmp( module->clazz->module_name, "cff" ) )
           {
-            FT_UInt  new_cff_hinting_engine;
-
-
-            new_cff_hinting_engine =
-              ( column->cff_hinting_engine + 1 ) % HINTING_ENGINE_MAX;
-
-            error = FT_Property_Set( state->library,
-                                     "cff",
-                                     "hinting-engine",
-                                     &new_cff_hinting_engine );
-            if ( !error )
-              column->cff_hinting_engine = new_cff_hinting_engine;
+            FTDemo_Event_Cff_Hinting_Engine_Change(
+              state->library,
+              &column->cff_hinting_engine,
+              1 );
+          }
+          else if ( !strcmp( module->clazz->module_name, "type1" ) )
+          {
+            FTDemo_Event_Type1_Hinting_Engine_Change(
+              state->library,
+              &column->type1_hinting_engine,
+              1 );
+          }
+          else if ( !strcmp( module->clazz->module_name, "t1cid" ) )
+          {
+            FTDemo_Event_T1cid_Hinting_Engine_Change(
+              state->library,
+              &column->t1cid_hinting_engine,
+              1 );
           }
           else if ( !strcmp( module->clazz->module_name, "truetype" ) )
           {
@@ -1455,8 +1508,8 @@
     ADisplayRec     adisplay[1];
     RenderStateRec  state[1];
     DisplayRec      display[1];
-    int             width      = DIM_X;
-    int             height     = DIM_Y;
+    int             width      = 640;
+    int             height     = 480;
     int             resolution = -1;
     double          size       = -1;
     const char*     textfile   = NULL;
@@ -1550,10 +1603,15 @@
         fseek( tfile, 0, SEEK_SET );
         text = (char*)malloc( (unsigned int)( tsize + 1 ) );
 
-        if ( text != NULL )
+        if ( text )
         {
-          fread( text, (unsigned int)tsize, 1, tfile );
-          text[tsize] = 0;
+          if ( !fread( text, (unsigned int)tsize, 1, tfile ) )
+          {
+            fprintf( stderr, "read error\n" );
+            text = (char *)default_text;
+          }
+          else
+            text[tsize] = '\0';
         }
         else
         {
