@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-2018 by                                                  */
+/*  Copyright (C) 1996-2019 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -95,8 +95,7 @@
 
   typedef struct  GridStatusRec_
   {
-    int          width;
-    int          height;
+    const char*  dims;
 
     int          ptsize;
     int          res;
@@ -159,12 +158,14 @@
 
   static GridStatusRec  status;
 
+  static FT_Glyph  circle;
+
 
   static void
   grid_status_init( GridStatus  st )
   {
-    st->width         = DIM_X;
-    st->height        = DIM_Y;
+    st->dims          = DIM;
+    st->res           = 72;
 
     st->scale         = 64;
     st->x_origin      = 0;
@@ -235,26 +236,19 @@
   {
     FT_Size     size;
     FT_Error    err    = FTDemo_Get_Size( handle, &size );
-    FT_F26Dot6  margin = 4;
+    FT_F26Dot6  margin = 6;
 
 
     if ( !err )
     {
-      FT_Face  face = size->face;
-
-      int  xmin = FT_MulFix( face->bbox.xMin, size->metrics.x_scale );
-      int  ymin = FT_MulFix( face->bbox.yMin, size->metrics.y_scale );
-      int  xmax = FT_MulFix( face->bbox.xMax, size->metrics.x_scale );
-      int  ymax = FT_MulFix( face->bbox.yMax, size->metrics.y_scale );
+      int  xmin = 0;
+      int  ymin = size->metrics.descender;
+      int  xmax = size->metrics.max_advance;
+      int  ymax = size->metrics.ascender;
 
       FT_F26Dot6  x_scale, y_scale;
 
 
-      xmin &= ~63;
-      ymin &= ~63;
-      xmax  = ( xmax + 63 ) & ~63;
-      ymax  = ( ymax + 63 ) & ~63;
-
       if ( xmax - xmin )
         x_scale = st->disp_width  * ( 64 - 2 * margin ) / ( xmax - xmin );
       else
@@ -270,8 +264,8 @@
       else
         st->scale = y_scale;
 
-      st->x_origin = st->disp_width  * margin          - xmin * st->scale;
-      st->y_origin = st->disp_height * ( 64 - margin ) + ymin * st->scale;
+      st->x_origin = 32 * st->disp_width  - ( xmax + xmin ) * st->scale / 2;
+      st->y_origin = 32 * st->disp_height + ( ymax + ymin ) * st->scale / 2;
     }
     else
     {
@@ -405,128 +399,20 @@
 
 
   static void
-  ft_bitmap_draw( FT_Bitmap*       bitmap,
-                  int              x,
-                  int              y,
-                  FTDemo_Display*  display,
-                  grColor          color )
+  circle_init( FTDemo_Handle*  handle,
+               FT_F26Dot6      radius )
   {
-    grBitmap  gbit;
-
-
-    gbit.width  = (int)bitmap->width;
-    gbit.rows   = (int)bitmap->rows;
-    gbit.pitch  = bitmap->pitch;
-    gbit.buffer = bitmap->buffer;
-
-    switch ( bitmap->pixel_mode )
-    {
-    case FT_PIXEL_MODE_GRAY:
-      gbit.mode  = gr_pixel_mode_gray;
-      gbit.grays = 256;
-      break;
-
-    case FT_PIXEL_MODE_MONO:
-      gbit.mode  = gr_pixel_mode_mono;
-      gbit.grays = 2;
-      break;
-
-    case FT_PIXEL_MODE_LCD:
-      gbit.mode  = gr_pixel_mode_lcd;
-      gbit.grays = 256;
-      break;
-
-    case FT_PIXEL_MODE_LCD_V:
-      gbit.mode  = gr_pixel_mode_lcdv;
-      gbit.grays = 256;
-      break;
-
-    default:
-      return;
-    }
-
-    grBlitGlyphToBitmap( display->bitmap, &gbit, x, y, color );
-  }
-
-
-  static void
-  ft_outline_draw( FT_Outline*      outline,
-                   double           scale,
-                   int              pen_x,
-                   int              pen_y,
-                   FTDemo_Handle*   handle,
-                   FTDemo_Display*  display,
-                   grColor          color )
-  {
-    FT_Outline  transformed;
-    FT_BBox     cbox;
-    FT_Bitmap   bitm;
-
-
-    FT_Outline_New( handle->library,
-                    (FT_UInt)outline->n_points,
-                    outline->n_contours,
-                    &transformed );
-
-    FT_Outline_Copy( outline, &transformed );
-
-    if ( scale != 1. )
-    {
-      int  nn;
-
-
-      for ( nn = 0; nn < transformed.n_points; nn++ )
-      {
-        FT_Vector*  vec = &transformed.points[nn];
-
-
-        vec->x = (FT_F26Dot6)( vec->x * scale );
-        vec->y = (FT_F26Dot6)( vec->y * scale );
-      }
-    }
-
-    FT_Outline_Get_CBox( &transformed, &cbox );
-    cbox.xMin &= ~63;
-    cbox.yMin &= ~63;
-    cbox.xMax  = ( cbox.xMax + 63 ) & ~63;
-    cbox.yMax  = ( cbox.yMax + 63 ) & ~63;
-
-    bitm.width      = (unsigned int)( ( cbox.xMax - cbox.xMin ) >> 6 );
-    bitm.rows       = (unsigned int)( ( cbox.yMax - cbox.yMin ) >> 6 );
-    bitm.pitch      = (int)bitm.width;
-    bitm.num_grays  = 256;
-    bitm.pixel_mode = FT_PIXEL_MODE_GRAY;
-    bitm.buffer     = (unsigned char*)calloc( (unsigned int)bitm.pitch,
-                                              bitm.rows );
-
-    FT_Outline_Translate( &transformed, -cbox.xMin, -cbox.yMin );
-    FT_Outline_Get_Bitmap( handle->library, &transformed, &bitm );
-
-    ft_bitmap_draw( &bitm,
-                    pen_x + ( cbox.xMin >> 6 ),
-                    pen_y - ( cbox.yMax >> 6 ),
-                    display,
-                    color );
-
-    free( bitm.buffer );
-    FT_Outline_Done( handle->library, &transformed );
-  }
-
-
-  static void
-  ft_outline_new_circle( FT_Outline*     outline,
-                         FT_F26Dot6      radius,
-                         FTDemo_Handle*  handle )
-  {
-    char*       tag;
-    FT_Vector*  vec;
-    FT_F26Dot6  disp = (FT_F26Dot6)( radius * 0.5523 );
+    FT_Outline*  outline;
+    char*        tag;
+    FT_Vector*   vec;
+    FT_F26Dot6   disp = (FT_F26Dot6)( radius * 0.5523 );
     /* so that Bézier curve touches circle at 0, 45, and 90 degrees */
 
 
+    FT_New_Glyph( handle->library, FT_GLYPH_FORMAT_OUTLINE, &circle );
+
+    outline = &((FT_OutlineGlyph)circle)->outline;
     FT_Outline_New( handle->library, 12, 1, outline );
-    outline->n_points    = 12;
-    outline->n_contours  = 1;
     outline->contours[0] = outline->n_points - 1;
 
     vec = outline->points;
@@ -550,22 +436,21 @@
   static void
   circle_draw( FT_F26Dot6       center_x,
                FT_F26Dot6       center_y,
-               FT_F26Dot6       radius,
                FTDemo_Handle*   handle,
                FTDemo_Display*  display,
                grColor          color )
   {
-    FT_Outline  outline;
+    FT_Outline*  outline = &((FT_OutlineGlyph)circle)->outline;
+    int  x = center_x >> 6;
+    int  y = center_y >> 6;
 
 
-    ft_outline_new_circle( &outline, radius, handle );
     /* subpixel adjustment considering downward direction of y-axis */
-    FT_Outline_Translate( &outline, center_x & 63, -( center_y & 63 ) );
+    FT_Outline_Translate( outline, center_x & 63, -( center_y & 63 ) );
 
-    ft_outline_draw( &outline, 1., ( center_x >> 6 ), ( center_y >> 6 ),
-                     handle, display, color );
+    FTDemo_Draw_Glyph_Color( handle, display, circle, &x, &y, color );
 
-    FT_Outline_Done( handle->library, &outline );
+    FT_Outline_Translate( outline, -( center_x & 63 ), center_y & 63 );
   }
 
 
@@ -605,6 +490,11 @@
           for ( k = 1; k < scale; k++, line += pitch * scale )
             memcpy( line + pitch * scale, line, (size_t)( pitch * scale ) );
           line += pitch * scale;
+
+          /* center specks */
+          if ( scale > 8 )
+            for ( j = scale / 2; j < width * scale; j += scale )
+              line[j / 8 - scale / 2 * pitch * scale] ^= 0x80 >> ( j & 7 );
         }
         break;
 
@@ -689,46 +579,28 @@
     if ( err )
       return;
 
+    glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)st->Num );
+
 #ifdef FT_DEBUG_AUTOFIT
-    /* Draw segment before drawing glyph. */
-    if ( status.do_segment )
-    {
-      /* Force hinting first in order to collect segment info. */
-      _af_debug_disable_horz_hints = 0;
-      _af_debug_disable_vert_hints = 0;
-
-      if ( !FT_Load_Glyph( size->face, (FT_UInt)st->Num,
-                           FT_LOAD_DEFAULT        |
-                           FT_LOAD_NO_BITMAP      |
-                           FT_LOAD_FORCE_AUTOHINT |
-                           FT_LOAD_TARGET_NORMAL ) )
-        grid_hint_draw_segment( &status, size, _af_debug_hints );
-    }
-
     _af_debug_disable_horz_hints = !st->do_horz_hints;
     _af_debug_disable_vert_hints = !st->do_vert_hints;
     _af_debug_disable_blue_hints = !st->do_blue_hints;
 #endif
 
-    if ( handle->encoding == FT_ENCODING_ORDER )
-      glyph_idx = (FT_UInt)st->Num;
-    else
-      glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)st->Num );
-
-
     if ( FT_Load_Glyph( size->face, glyph_idx,
                         handle->load_flags | FT_LOAD_NO_BITMAP ) )
       return;
 
+    slot = size->face->glyph;
+
     if ( st->do_grid )
     {
       /* show advance width */
       grFillVLine( st->disp_bitmap,
                    st->x_origin +
-                     ( ( size->face->glyph->metrics.horiAdvance +
-                         size->face->glyph->lsb_delta           -
-                         size->face->glyph->rsb_delta           ) *
-                       scale >> 6 ),
+                     ( ( slot->metrics.horiAdvance +
+                         slot->lsb_delta           -
+                         slot->rsb_delta           ) * scale >> 6 ),
                    0,
                    st->disp_height,
                    st->axis_color );
@@ -746,7 +618,35 @@
                    st->axis_color );
     }
 
-    slot = size->face->glyph;
+    /* render scaled bitmap */
+    if ( st->work & DO_BITMAP )
+    {
+      FT_Glyph        glyph, glyf;
+      int             left, top, x_advance, y_advance;
+      grBitmap        bitg;
+
+
+      FT_Get_Glyph( slot, &glyph );
+      error  = FTDemo_Glyph_To_Bitmap( handle, glyph, &bitg, &left, &top,
+                                       &x_advance, &y_advance, &glyf);
+
+      if ( !error )
+      {
+        bitmap_scale( &bitg, scale );
+
+        grBlitGlyphToBitmap( display->bitmap, &bitg,
+                             ox + left * scale, oy - top * scale,
+                             st->axis_color );
+
+        free( bitg.buffer );
+
+        if ( glyf )
+          FT_Done_Glyph( glyf );
+      }
+
+      FT_Done_Glyph( glyph );
+    }
+
     if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
     {
       FT_Glyph     glyph;
@@ -754,34 +654,11 @@
       int          nn;
 
 
-      /* render scaled bitmap */
-      if ( st->work & DO_BITMAP )
-      {
-        int             left, top, x_advance, y_advance;
-        grBitmap        bitg;
-        FT_Glyph        glyf;
-
-
-        FT_Get_Glyph( slot, &glyph );
-        error  = FTDemo_Glyph_To_Bitmap( handle, glyph, &bitg, &left, &top,
-                                         &x_advance, &y_advance, &glyf);
-
-        if ( !error )
-        {
-          bitmap_scale( &bitg, scale );
-
-          grBlitGlyphToBitmap( display->bitmap, &bitg,
-                               ox + left * scale, oy - top * scale,
-                               st->axis_color );
-
-          free( bitg.buffer );
-
-          if ( glyf )
-            FT_Done_Glyph( glyf );
-        }
-
-        FT_Done_Glyph( glyph );
-      }
+#ifdef FT_DEBUG_AUTOFIT
+      /* Draw segment before drawing glyph. */
+      if ( status.do_segment && handle->load_flags & FT_LOAD_FORCE_AUTOHINT )
+        grid_hint_draw_segment( &status, size, _af_debug_hints );
+#endif
 
       /* scale the outline */
       for ( nn = 0; nn < gimage->n_points; nn++ )
@@ -813,7 +690,6 @@
           circle_draw(
             st->x_origin * 64 + gimage->points[nn].x,
             st->y_origin * 64 - gimage->points[nn].y,
-            128,
             handle,
             display,
             ( gimage->tags[nn] & FT_CURVE_TAG_ON ) ? st->on_color
@@ -1854,17 +1730,14 @@
       "            `.afm' or `.pfm').\n"
       "\n" );
     fprintf( stderr,
-      "  -w W      Set the window width to W pixels (default: %dpx).\n"
-      "  -h H      Set the window height to H pixels (default: %dpx).\n"
-      "\n",
-             DIM_X, DIM_Y );
-    fprintf( stderr,
+      "  -d WxHxD  Set the window width, height, and color depth\n"
+      "            (default: 640x480x24).\n"
       "  -r R      Use resolution R dpi (default: 72dpi).\n"
       "  -f index  Specify first index to display (default: 0).\n"
       "  -e enc    Specify encoding tag (default: no encoding).\n"
       "            Common values: `unic' (Unicode), `symb' (symbol),\n"
       "            `ADOB' (Adobe standard), `ADBC' (Adobe custom).\n"
-      "  -d \"axis1 axis2 ...\"\n"
+      "  -a \"axis1 axis2 ...\"\n"
       "            Specify the design coordinates for each\n"
       "            Multiple Master axis at start-up.  Implies `-n'.\n"
       "  -n        Don't display named instances of variation fonts.\n"
@@ -1888,14 +1761,14 @@
 
     while ( 1 )
     {
-      option = getopt( *argc, *argv, "d:e:f:h:nr:vw:" );
+      option = getopt( *argc, *argv, "a:d:e:f:nr:v" );
 
       if ( option == -1 )
         break;
 
       switch ( option )
       {
-      case 'd':
+      case 'a':
         {
           FT_UInt    cnt;
           FT_Fixed*  pos = status.requested_pos;
@@ -1915,6 +1788,10 @@
         }
         break;
 
+      case 'd':
+        status.dims = optarg;
+        break;
+
       case 'e':
         handle->encoding = FTDemo_Make_Encoding_Tag( optarg );
         status.Num       = 0x20;
@@ -1924,12 +1801,6 @@
         status.Num = atoi( optarg );
         break;
 
-      case 'h':
-        status.height = atoi( optarg );
-        if ( status.height < 1 )
-          usage( execname );
-        break;
-
       case 'n':
         status.no_named_instances = 1;
         break;
@@ -1955,12 +1826,6 @@
         }
         /* break; */
 
-      case 'w':
-        status.width = atoi( optarg );
-        if ( status.width < 1 )
-          usage( execname );
-        break;
-
       default:
         usage( execname );
         break;
@@ -1977,9 +1842,6 @@
     if ( status.ptsize == 0 )
       status.ptsize = 64 * 10;
 
-    if ( status.res <= 0 )
-      status.res = 72;
-
     (*argc)--;
     (*argv)++;
   }
@@ -2001,6 +1863,7 @@
     handle = FTDemo_New();
 
     grid_status_init( &status );
+    circle_init( handle, 128 );
     parse_cmdline( &argc, &argv );
 
     /* get the default value as compiled into FreeType */
@@ -2045,28 +1908,22 @@
     FT_Stroker_Set( status.stroker, 32, FT_STROKER_LINECAP_BUTT,
                       FT_STROKER_LINEJOIN_BEVEL, 0x20000 );
 
-    display = FTDemo_Display_New( gr_pixel_mode_rgb24,
-                                  status.width, status.height );
+    for ( ; argc > 0; argc--, argv++ )
+      FTDemo_Install_Font( handle, argv[0], 0,
+                           status.no_named_instances ? 1 : 0 );
+
+    if ( handle->num_fonts == 0 )
+      Fatal( "could not find/open any font file" );
+
+    display = FTDemo_Display_New( status.dims );
     if ( !display )
       Fatal( "could not allocate display surface" );
 
-    grid_status_display( &status, display );
-    grid_status_colors(  &status, display );
-
     grSetTitle( display->surface,
                 "FreeType Glyph Grid Viewer - press ? for help" );
 
-    for ( ; argc > 0; argc--, argv++ )
-    {
-      error = FTDemo_Install_Font( handle, argv[0], 1,
-                                   status.no_named_instances ? 1 : 0 );
-      if ( error == FT_Err_Invalid_Argument )
-        fprintf( stderr, "skipping font `%s' without outlines\n",
-                         argv[0] );
-    }
-
-    if ( handle->num_fonts == 0 )
-      Fatal( "could not find/open any font file" );
+    grid_status_display( &status, display );
+    grid_status_colors(  &status, display );
 
     event_font_change( 0 );
 
