@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-2018 by                                                  */
+/*  Copyright (C) 1996-2019 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -22,6 +22,7 @@
 #include <math.h>
 
 #include FT_LCD_FILTER_H
+#include FT_TRIGONOMETRY_H
 
 #define CELLSTRING_HEIGHT  8
 #define MAXPTSIZE          500   /* dtp */
@@ -74,14 +75,15 @@
   enum
   {
     RENDER_MODE_STRING,
+    RENDER_MODE_TEXT,
+    RENDER_MODE_WATERFALL,
     RENDER_MODE_KERNCMP,
     N_RENDER_MODES
   };
 
   static struct  status_
   {
-    int  width;
-    int  height;
+    const char*    dims;
 
     int            render_mode;
     unsigned long  encoding;
@@ -92,20 +94,78 @@
 
     FTDemo_String_Context  sc;
 
-    FT_Byte    gamma_ramp[256];   /* for show only */
     FT_Matrix  trans_matrix;
     int        font_index;
     char*      header;
     char       header_buffer[256];
 
-  } status = { DIM_X, DIM_Y,
-               RENDER_MODE_STRING, FT_ENCODING_UNICODE, 72, 48, 0, NULL,
-               { 0, 0, 0, 0, NULL },
-               { 0 }, { 0, 0, 0, 0 }, 0, NULL, { 0 } };
+  } status = { DIM, RENDER_MODE_STRING, FT_ENCODING_UNICODE, 72, 48, 0, NULL,
+               { 0, 0, 0x8000, 0, NULL, 0, 0 },
+               { 0, 0, 0, 0 }, 0, NULL, { 0 } };
 
   static FTDemo_Display*  display;
   static FTDemo_Handle*   handle;
 
+  static FT_Glyph  wheel, daisy, aster;  /* stress test glyphs */
+
+
+  static void
+  flower_init( FT_Glyph*     glyph,
+               FT_F26Dot6    radius,
+               unsigned int  i,
+               int           v,
+               int           w,
+               int           reflect,
+               char          order )
+  {
+    FT_Outline*  outline;
+    FT_Vector*   vec;
+    FT_Vector*   limit;
+    char*        tag;
+    FT_Fixed     s = FT_Sin( FT_ANGLE_PI4 / i );
+    FT_Pos       b, d, p = 0, q = radius;
+
+
+    FT_New_Glyph( handle->library, FT_GLYPH_FORMAT_OUTLINE, glyph );
+
+    outline = &((FT_OutlineGlyph)*glyph)->outline;
+
+    FT_Outline_New( handle->library, 6 * i, 1, outline );
+    outline->contours[0] = outline->n_points - 1;
+
+    if ( order == FT_CURVE_TAG_CUBIC )
+      q += q / 3;
+
+    for ( vec = outline->points, tag = outline->tags;
+          i--;
+          vec += 6,              tag += 6 )
+    {
+       b = p + FT_MulFix( q, s );
+       d = q - FT_MulFix( p, s );
+       vec[0].x = vec[3].y = 0;
+       vec[0].y = vec[3].x = 0;
+       tag[0]   = tag[3]   = FT_CURVE_TAG_ON;
+       vec[1].x = vec[5].y = ( v * p + w * b ) / ( v + w );
+       vec[1].y = vec[5].x = ( v * q + w * d ) / ( v + w );
+       tag[1]   = tag[5]   = order;
+       vec[2].x = vec[4].y = ( v * b + w * p ) / ( v + w );
+       vec[2].y = vec[4].x = ( v * d + w * q ) / ( v + w );
+       tag[2]   = tag[4]   = order;
+       p = b;
+       q = d;
+    }
+
+    limit = outline->points + outline->n_points;
+
+    if ( reflect & 1 )
+      for( vec = outline->points; vec < limit; vec++ )
+        vec->x = -vec->x;
+
+    if ( reflect & 2 )
+      for( vec = outline->points; vec < limit; vec++ )
+        vec->y = -vec->y;
+  }
+
 
   /*************************************************************************/
   /*************************************************************************/
@@ -157,7 +217,7 @@
     grWriteln( "  f         : toggle forced auto-hinting" );
     grWriteln( "  h         : toggle outline hinting" );
     grLn();
-    grWriteln( "  1-2       : select rendering mode" );
+    grWriteln( "  1-4       : select rendering mode" );
     grWriteln( "  l         : cycle through anti-aliasing modes" );
     grWriteln( "  k         : cycle through kerning modes" );
     grWriteln( "  t         : cycle through kerning degrees" );
@@ -176,8 +236,13 @@
     grWriteln( "  Page Up   : increase pointsize by 10 units" );
     grWriteln( "  Page Down : decrease pointsize by 10 units" );
     grLn();
-    grWriteln( "  Right     : rotate counter-clockwise" );
-    grWriteln( "  Left      : rotate clockwise" );
+    grWriteln( "  Left      : move left" );
+    grWriteln( "  Right     : move right" );
+    grWriteln( "  Home      : flush left" );
+    grWriteln( "  End       : flush right" );
+    grLn();
+    grWriteln( "  F5        : rotate counter-clockwise" );
+    grWriteln( "  F6        : rotate clockwise" );
     grWriteln( "  F7        : big rotate counter-clockwise" );
     grWriteln( "  F8        : big rotate clockwise" );
     grLn();
@@ -199,9 +264,18 @@
 
     FTDemo_Set_Current_Font( handle, handle->fonts[status.font_index] );
     FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
-    FTDemo_Update_Current_Flags( handle );
+  }
 
-    FTDemo_String_Set( handle, status.text );
+
+  static void
+  event_center_change( FT_Fixed  delta )
+  {
+    status.sc.center += delta;
+
+    if ( status.sc.center > 0x10000 )
+      status.sc.center = 0x10000;
+    else if ( status.sc.center < 0 )
+      status.sc.center = 0;
   }
 
 
@@ -292,6 +366,7 @@
 
     display->back_color = grFindColor( display->bitmap,  r,  g,  b, 0xff );
     display->fore_color = grFindColor( display->bitmap, ~r, ~g, ~b, 0xff );
+    display->warn_color = grFindColor( display->bitmap,  r, ~g, ~b, 0xff );
 
     i++;
   }
@@ -312,9 +387,6 @@
   static void
   event_gamma_change( double  delta )
   {
-    int     i;
-
-
     display->gamma += delta;
 
     if ( display->gamma > 3.0 )
@@ -323,10 +395,6 @@
       display->gamma = 0.1;
 
     grSetGlyphGamma( display->gamma );
-
-    for ( i = 0; i < 256; i++ )
-      status.gamma_ramp[i] = (FT_Byte)( pow( (double)i / 255., display->gamma )
-                                        * 255. + 0.5 );
   }
 
 
@@ -357,13 +425,13 @@
 
     switch ( status.render_mode )
     {
-    case RENDER_MODE_STRING:
-      status.header = NULL;
-      break;
-
     case RENDER_MODE_KERNCMP:
       status.header = (char *)"Kerning comparison";
       break;
+
+    default:
+      status.header = NULL;
+      break;
     }
   }
 
@@ -377,7 +445,7 @@
 
     if ( event->key >= '1' && event->key < '1' + N_RENDER_MODES )
     {
-      status.render_mode = event->key - '1';
+      status.render_mode = (int)( event->key - '1' );
       event_render_mode_change( 0 );
 
       return ret;
@@ -388,45 +456,37 @@
     case grKeyEsc:
     case grKEY( 'q' ):
       ret = 1;
-      break;
+      goto Exit;;
 
     case grKeyF1:
     case grKEY( '?' ):
       event_help();
-      break;
+      goto Exit;
 
     case grKEY( 'b' ):
       handle->use_sbits = !handle->use_sbits;
       status.header     = handle->use_sbits
                           ? (char *)"embedded bitmaps are now used when available"
                           : (char *)"embedded bitmaps are now ignored";
-
-      FTDemo_Update_Current_Flags( handle );
-      break;
+      goto Flags;
 
     case grKEY( 'f' ):
       handle->autohint = !handle->autohint;
       status.header     = handle->autohint
                           ? (char *)"forced auto-hinting is now on"
                           : (char *)"forced auto-hinting is now off";
-
-      FTDemo_Update_Current_Flags( handle );
-      break;
+      goto Flags;
 
     case grKEY( 'h' ):
       handle->hinted = !handle->hinted;
       status.header   = handle->hinted
                         ? (char *)"glyph hinting is now active"
                         : (char *)"glyph hinting is now ignored";
-
-      FTDemo_Update_Current_Flags( handle );
-      break;
+      goto Flags;
 
     case grKEY( 'l' ):
       event_lcdmode_change();
-
-      FTDemo_Update_Current_Flags( handle );
-      break;
+      goto Flags;
 
     case grKEY( 'k' ):
       sc->kerning_mode = ( sc->kerning_mode + 1 ) % N_KERNING_MODES;
@@ -436,7 +496,7 @@
         : sc->kerning_mode == KERNING_MODE_NORMAL
           ? (char *)"pair kerning is now active"
           : (char *)"pair kerning is now ignored";
-      break;
+      goto String;
 
     case grKEY( 't' ):
       sc->kerning_degree = ( sc->kerning_degree + 1 ) % N_KERNING_DEGREES;
@@ -448,78 +508,72 @@
           : sc->kerning_degree == KERNING_DEGREE_MEDIUM
             ? (char *)"medium track kerning active"
             : (char *)"tight track kerning active";
-      break;
+      goto String;
 
     case grKeySpace:
       event_color_change();
-      break;
+      goto Exit;
 
     case grKeyTab:
       event_text_change();
       FTDemo_String_Set( handle, status.text );
-      break;
+      goto String;
 
     case grKEY( 'V' ):
       sc->vertical  = !sc->vertical;
       status.header = sc->vertical
                       ? (char *)"using vertical layout"
                       : (char *)"using horizontal layout";
-      break;
+      goto Exit;
 
     case grKEY( 'g' ):
       event_gamma_change( 0.1 );
-      break;
+      goto Exit;
 
     case grKEY( 'v' ):
       event_gamma_change( -0.1 );
-      break;
+      goto Exit;
 
     case grKEY( 'n' ):
       event_font_change( 1 );
-      break;
+      FTDemo_String_Set( handle, status.text );
+      goto Flags;
 
     case grKEY( 'p' ):
       event_font_change( -1 );
-      break;
+      FTDemo_String_Set( handle, status.text );
+      goto Flags;
 
-    case grKeyUp:       event_size_change(   64 ); break;
-    case grKeyDown:     event_size_change(  -64 ); break;
-    case grKeyPageUp:   event_size_change(  640 ); break;
-    case grKeyPageDown: event_size_change( -640 ); break;
+    case grKeyUp:       event_size_change(   64 ); goto String;
+    case grKeyDown:     event_size_change(  -64 ); goto String;
+    case grKeyPageUp:   event_size_change(  640 ); goto String;
+    case grKeyPageDown: event_size_change( -640 ); goto String;
 
-    case grKeyLeft:  event_angle_change(    -3 ); break;
-    case grKeyRight: event_angle_change(     3 ); break;
-    case grKeyF7:    event_angle_change(   -30 ); break;
-    case grKeyF8:    event_angle_change(    30 ); break;
+    case grKeyLeft:  event_center_change( -0x800 ); goto Exit;
+    case grKeyRight: event_center_change(  0x800 ); goto Exit;
+    case grKeyHome:  event_center_change( -0x10000 ); goto Exit;
+    case grKeyEnd:   event_center_change(  0x10000 ); goto Exit;
+
+    case grKeyF5: event_angle_change(  -3 ); goto Exit;
+    case grKeyF6: event_angle_change(   3 ); goto Exit;
+    case grKeyF7: event_angle_change( -30 ); goto Exit;
+    case grKeyF8: event_angle_change(  30 ); goto Exit;
 
     default:
       break;
     }
 
+  Flags:
+    FTDemo_Update_Current_Flags( handle );
+
+  String:
+    FTDemo_String_Load( handle, &status.sc );
+
+  Exit:
     return ret;
   }
 
 
-  static void
-  gamma_ramp_draw( FT_Byte    gamma_ramp[256],
-                   grBitmap*  bitmap )
-  {
-    int       i, x, y;
-    int       bpp = bitmap->pitch / bitmap->width;
-    FT_Byte*  p = (FT_Byte*)bitmap->buffer;
-
-
-    if ( bitmap->pitch < 0 )
-      p += -bitmap->pitch * ( bitmap->rows - 1 );
-
-    x = ( bitmap->width - 256 ) / 2;
-    y = ( bitmap->rows + 256 ) / 2;
-
-    for (i = 0; i < 256; i++)
-      p[bitmap->pitch * ( y - i ) + bpp * ( x + gamma_ramp[i] )] ^= 0xFF;
-  }
-
-
   static void
   write_header( FT_Error  error_code )
   {
@@ -557,13 +611,10 @@
       "            `.afm' or `.pfm').\n"
       "\n" );
     fprintf( stderr,
-      "  -w W      Set the window width to W pixels (default: %dpx).\n"
-      "  -h H      Set the window height to H pixels (default: %dpx).\n"
-      "\n",
-             DIM_X, DIM_Y );
-    fprintf( stderr,
+      "  -d WxHxD  Set the window width, height, and color depth\n"
+      "            (default: 640x480x24).\n"
       "  -r R      Use resolution R dpi (default: 72dpi).\n"
-      "  -e enc    Specify encoding tag (default: no encoding).\n"
+      "  -e enc    Specify encoding tag (default: Unicode).\n"
       "            Common values: `unic' (Unicode), `symb' (symbol),\n"
       "            `ADOB' (Adobe standard), `ADBC' (Adobe custom).\n"
       "  -m text   Use `text' for rendering.\n"
@@ -587,21 +638,19 @@
 
     while ( 1 )
     {
-      option = getopt( *argc, *argv, "e:h:m:r:vw:" );
+      option = getopt( *argc, *argv, "d:e:m:r:v" );
 
       if ( option == -1 )
         break;
 
       switch ( option )
       {
-      case 'e':
-        status.encoding = FTDemo_Make_Encoding_Tag( optarg );
+      case 'd':
+        status.dims = optarg;
         break;
 
-      case 'h':
-        status.height = atoi( optarg );
-        if ( status.height < 1 )
-          usage( execname );
+      case 'e':
+        status.encoding = FTDemo_Make_Encoding_Tag( optarg );
         break;
 
       case 'm':
@@ -631,12 +680,6 @@
         }
         /* break; */
 
-      case 'w':
-        status.width = atoi( optarg );
-        if ( status.width < 1 )
-          usage( execname );
-        break;
-
       default:
         usage( execname );
         break;
@@ -658,6 +701,164 @@
   }
 
 
+  static FT_Error
+  Render_String( void )
+  {
+    int x, y = display->bitmap->rows - 4;
+
+
+    x = 4;
+    FTDemo_Draw_Glyph( handle, display, daisy, &x, &y );
+
+    x = display->bitmap->width - 4;
+    FTDemo_Draw_Glyph( handle, display, aster, &x, &y );
+
+    FTDemo_String_Draw( handle, display,
+                        &status.sc,
+                        FT_MulFix( display->bitmap->width, status.sc.center),
+                        display->bitmap->rows / 2 );
+
+    return FT_Err_Ok;
+  }
+
+
+  static FT_Error
+  Render_Text( void )
+  {
+    int      x = FT_MulFix( display->bitmap->width, status.sc.center);
+    int      y, step_y;
+    int      offset = 0;
+    FT_Size  size;
+
+    FTDemo_String_Context sc = status.sc;
+
+
+    sc.extent   = display->bitmap->width * 64;
+    sc.vertical = 0;
+
+    error = FTDemo_Get_Size( handle, &size );
+    if ( error )
+      return error;
+
+    step_y = ( size->metrics.height >> 6 ) + 1;
+    y      = 40 + ( size->metrics.ascender >> 6 );
+
+    for ( ; y < display->bitmap->rows + ( size->metrics.descender >> 6 );
+            y += step_y )
+    {
+      sc.offset = offset;
+
+      offset += FTDemo_String_Draw( handle, display, &sc, x, y );
+
+      offset %= handle->string_length;
+    }
+
+    return FT_Err_Ok;
+  }
+
+
+  static FT_Error
+  Render_Waterfall( void )
+  {
+    int      pt_size = status.ptsize, step, pt_height;
+    int      y = 40;
+    int      x = FT_MulFix( display->bitmap->width, status.sc.center);
+    FT_Size  size;
+
+    FTDemo_String_Context sc = status.sc;
+
+
+    sc.vertical = 0;
+
+    pt_height = 64 * 72 * display->bitmap->rows / status.res;
+    step      = ( pt_size * pt_size / pt_height + 64 ) & ~63;
+    pt_size   = pt_size - step * ( pt_size / step ); /* remainder */
+
+    while ( 1 )
+    {
+      pt_size += step;
+
+      FTDemo_Set_Current_Charsize( handle, pt_size, status.res );
+      FTDemo_String_Load( handle, &status.sc );
+
+      error = FTDemo_Get_Size( handle, &size );
+      if ( error )
+      {
+        /* probably a non-existent bitmap font size */
+        continue;
+      }
+
+      if ( pt_size == status.ptsize )
+        grFillHLine( display->bitmap, x - 4, y, 8, display->warn_color );
+
+      y += ( size->metrics.height >> 6 ) + 1;
+
+      if ( y >= display->bitmap->rows )
+        break;
+
+      if ( pt_size == status.ptsize )
+        grFillHLine( display->bitmap, x - 4, y, 8, display->warn_color );
+
+      FTDemo_String_Draw( handle, display, &sc,
+                          x, y + ( size->metrics.descender >> 6 ) );
+    }
+
+    FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
+    FTDemo_String_Load( handle, &status.sc );
+
+    return FT_Err_Ok;
+  }
+
+
+  static FT_Error
+  Render_KernCmp( void )
+  {
+    FT_Size                size;
+    FTDemo_String_Context  sc = { 0, 0, 0, 0, NULL, 0, 0 };
+    FT_Int                 x, y;
+    FT_Int                 height;
+
+
+    x = 55;
+
+    FTDemo_Get_Size( handle, &size );
+    height = size->metrics.y_ppem;
+    if ( height < CELLSTRING_HEIGHT )
+      height = CELLSTRING_HEIGHT;
+
+    /* First line: none */
+    FTDemo_String_Load( handle, &sc );
+
+    y = CELLSTRING_HEIGHT * 2 + display->bitmap->rows / 4 + height;
+    grWriteCellString( display->bitmap, 5,
+                       y - ( height + CELLSTRING_HEIGHT ) / 2,
+                       "none", display->fore_color );
+    FTDemo_String_Draw( handle, display, &sc, x, y );
+
+    /* Second line: track kern only */
+    sc.kerning_degree = status.sc.kerning_degree;
+    FTDemo_String_Load( handle, &sc );
+
+    y += height;
+    grWriteCellString( display->bitmap, 5,
+                       y - ( height + CELLSTRING_HEIGHT ) / 2,
+                       "track", display->fore_color );
+    FTDemo_String_Draw( handle, display, &sc, x, y );
+
+    /* Third line: track kern + pair kern */
+    sc.kerning_mode = status.sc.kerning_mode;
+    FTDemo_String_Load( handle, &sc );
+
+    y += height;
+    grWriteCellString( display->bitmap, 5,
+                       y - ( height + CELLSTRING_HEIGHT ) / 2,
+                       "both", display->fore_color );
+    FTDemo_String_Draw( handle, display, &sc, x, y );
+
+    return FT_Err_Ok;
+  }
+
+
   int
   main( int     argc,
         char**  argv )
@@ -670,11 +871,13 @@
 
     parse_cmdline( &argc, &argv );
 
+    flower_init( &wheel, 8192, 20, 3, 1, 0, FT_CURVE_TAG_ON    );
+    flower_init( &daisy, 8192, 20, 4, 1, 0, FT_CURVE_TAG_CONIC );
+    flower_init( &aster, 8192, 20, 5, 1, 1, FT_CURVE_TAG_CUBIC );
+
     FT_Library_SetLcdFilter( handle->library, FT_LCD_FILTER_LIGHT );
 
     handle->encoding  = status.encoding;
-    handle->use_sbits = 0;
-    FTDemo_Update_Current_Flags( handle );
 
     for ( ; argc > 0; argc--, argv++ )
     {
@@ -693,8 +896,7 @@
     if ( handle->num_fonts == 0 )
       PanicZ( "could not open any font file" );
 
-    display = FTDemo_Display_New( gr_pixel_mode_rgb24,
-                                  status.width, status.height );
+    display = FTDemo_Display_New( status.dims );
 
     if ( !display )
       PanicZ( "could not allocate display surface" );
@@ -709,68 +911,32 @@
 
     event_color_change();
     event_gamma_change( 0 );
+
     event_font_change( 0 );
+    FTDemo_String_Set( handle, status.text );
+    FTDemo_Update_Current_Flags( handle );
+    FTDemo_String_Load( handle, &status.sc );
 
     do
     {
       FTDemo_Display_Clear( display );
 
-      gamma_ramp_draw( status.gamma_ramp, display->bitmap );
-
       switch ( status.render_mode )
       {
       case RENDER_MODE_STRING:
-        status.sc.center = 1L << 15;
-        error = FTDemo_String_Draw( handle, display,
-                                    &status.sc,
-                                    display->bitmap->width / 2,
-                                    display->bitmap->rows / 2 );
+        error = Render_String();
+        break;
+
+      case RENDER_MODE_TEXT:
+        error = Render_Text();
+        break;
+
+      case RENDER_MODE_WATERFALL:
+        error = Render_Waterfall();
         break;
 
       case RENDER_MODE_KERNCMP:
-        {
-          FTDemo_String_Context  sc = status.sc;
-          FT_Int                 x, y;
-          FT_Int                 height;
-
-
-          x = 55;
-
-          height = ( status.ptsize * status.res / 72 + 32 ) >> 6;
-          if ( height < CELLSTRING_HEIGHT )
-            height = CELLSTRING_HEIGHT;
-
-          /* First line: none */
-          sc.center         = 0;
-          sc.kerning_mode   = 0;
-          sc.kerning_degree = 0;
-          sc.vertical       = 0;
-          sc.matrix         = NULL;
-
-          y = CELLSTRING_HEIGHT * 2 + display->bitmap->rows / 4 + height;
-          grWriteCellString( display->bitmap, 5,
-                             y - ( height + CELLSTRING_HEIGHT ) / 2,
-                             "none", display->fore_color );
-          error = FTDemo_String_Draw( handle, display, &sc, x, y );
-
-          /* Second line: track kern only */
-          sc.kerning_degree = status.sc.kerning_degree;
-
-          y += height;
-          grWriteCellString( display->bitmap, 5,
-                             y - ( height + CELLSTRING_HEIGHT ) / 2,
-                             "track", display->fore_color );
-          error = FTDemo_String_Draw( handle, display, &sc, x, y );
-
-          /* Third line: track kern + pair kern */
-          sc.kerning_mode = status.sc.kerning_mode;
-
-          y += height;
-          grWriteCellString( display->bitmap, 5,
-                             y - ( height + CELLSTRING_HEIGHT ) / 2,
-                             "both", display->fore_color );
-          error = FTDemo_String_Draw( handle, display, &sc, x, y );
-        }
+        error = Render_KernCmp();
         break;
       }
 
